import _thread
from lib.httpserver import (HTTPServer, HTTPResponse)
from lib.httpserver.url import HTTPRequest
from lib.html_builder import HTMLBuilder
import src.data as data
from src.wireless import espnow_recv, espnow_send, espnow_send_raw


app = HTTPServer(timeout=10)

status = {
    "page_title": {
        "type": "item",
        "value": "TM-SIM Home"
    },
    "slave_selected": {
        "type": "item",
        "value": "Slave 1"
    },
    "select_slave": {
        "type": "radio",
        "value": [f'Slave {x}' for x in range(1, len(data.slave_list)+1)]
    },
    "to_slave": {
        "type": "radio",
        "value": [f'Slave {x}' for x in range(1, len(data.slave_list)+1)]
    },
    "rule_used": {
        "type": "item",
        "value": ""
    },
    "band_current": {
        "type": "item",
        "value": ""
    }
}


def unescape(string: str) -> str:
    # unescapes the chars needed for tms
    # namely ; , - _ < >
    return (string
        .replace('%3B',';')
        .replace('%2C',',')
        .replace('%2D','-')
        .replace('%5F','_')
        .replace('%3C','<')
        .replace('%3E','>')
        .replace('+', ' ')
    )


@app.route("GET", "/")
def root(con, _):
    global status

    HTTPResponse(con, 200, "HTML").send_raw(
        HTMLBuilder(["html/head.html","html/home.html"], status).build()
    )


@app.route("GET", "/docs")
def docs(con, _):
    HTTPResponse(con, 200, "HTML").send_file("html/docs.html")


@app.route("GET", "/definition")
def definition(con, request: HTTPRequest):
    # Get slave and tm definition from request and send off
    para_tm = request.parameters.get("tm")
    para_slave = request.parameters.get("select_slave")

    if para_tm != None and para_slave != None:
        tm = "DEFINE:" + unescape(str(para_tm))
        slave_num = int(str(para_slave)[6:]) -1
        slave = data.slave_list[slave_num]

        print(tm)
        espnow_send(slave, tm)

    HTTPResponse(con, 302).redirect("/")


@app.route("GET", "/input")
def input(con, request: HTTPRequest):
    para_band = request.parameters.get("band")
    para_ss = request.parameters.get("select_slave")

    if para_band != None and para_ss != None:
        # Get slave and input band from request and send off
        band = "BAND:" + str(para_band)
        slave_num = int(str(para_ss)[6:]) -1
        slave = data.slave_list[slave_num]

        espnow_send(slave, band)

    HTTPResponse(con, 302).redirect("/")


@app.route("GET", "/run")
def run(con, request: HTTPRequest):
    # use builder to incorporate that into response
    global status

    select_slave = request.parameters.get("select_slave")
    if select_slave != None:
        select_slave = str(select_slave)
        if "Slave" in select_slave:
            status["slave_selected"]["value"] = unescape(select_slave)

        # Go to simulator page where you can run all or step through
        HTTPResponse(con, 200, "HTML").send_raw(
            HTMLBuilder(["html/head.html", "html/run.html"], status).build()
        )

    HTTPResponse(con, 302).redirect("/")


@app.route("GET", "/run_all")
def run_all(con, _):
    global status
    # Run the tm on the current controller to completion
    slave_num = int(status["slave_selected"]["value"][6:]) -1
    espnow_send(data.slave_list[slave_num], "RUN")
    status["band_current"]["value"] = espnow_recv()

    HTTPResponse(con, 302).redirect("/run")


@app.route("GET", "/run_step")
def run_step(con, _):
    global status
    slave_num = int(status["slave_selected"]["value"][6:]) -1

    espnow_send(data.slave_list[slave_num], "STEP")
    rule_used, band_current = espnow_recv().split(':')

    status["rule_used"]["value"] = rule_used
    status["band_current"]["value"] = band_current

    HTTPResponse(con, 302).redirect("/run")


@app.route("GET", "/pass")
def pass_band(con, request: HTTPRequest):
    para_sfrom = request.parameters.get("select_slave")
    para_sto = request.parameters.get("to_slave")

    if para_sfrom == None or para_sto == None:
        HTTPResponse(con, 302).redirect("/")

    try:
        slave_from = int(str(para_sfrom)[6:]) -1
        slave_to = int(str(para_sto)[6:]) -1

        slave_from = data.slave_list[slave_from]
        slave_to = data.slave_list[slave_to]
        
        # Give PASS command and send peer to pass to
        espnow_send(slave_from, "PASS")
        espnow_send_raw(slave_from, slave_to)
    except:
        print(f"WEB:ERR:PASS:Slaves not selected")

    
    HTTPResponse(con, 302).redirect("/run")


@app.route("GET", "/result")
def result(con, _):
    # htmlbuilder dat hoe to show result of every tm 
    # and also endresult of computation chain
    result_list = []
    for i in range(0, len(data.slave_list)):
        # todo: parse result_dict of src.data to display res
        while not espnow_send(data.slave_list[i], "RESULT"):
            pass
        
        recv = espnow_recv().split(":")
        if len(recv) > 1:
            result_list.append(espnow_recv()[1])
        else:
            result_list.append(espnow_recv())
    
    results = {
            "results": {
            "type": "list",
            "value": result_list
        }
    }

    HTTPResponse(con, 200, "HTML").send_raw(
        HTMLBuilder(["html/head.html", "html/result.html"], results).build()
    )


@app.route("GET", "/reset")
def reset(con, _):
    global status 
    # rule_used and band_current are reset on slave side also
    status["rule_used"]["value"] = ""
    status["band_current"]["value"] = ""
    slave_num = int(status["slave_selected"]["value"][6:]) -1

    # send reset
    espnow_send(data.slave_list[slave_num], "RESET")

    HTTPResponse(con, 302).redirect("/run")


def start() -> None:
    _thread.start_new_thread(app.start, ())

